package ru.dbotthepony.mc.otm.config

import ru.dbotthepony.mc.otm.block.entity.matter.MatterBottlerBlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterReplicatorBlockEntity
import ru.dbotthepony.mc.otm.block.entity.tech.AndroidStationBlockEntity
import ru.dbotthepony.mc.otm.block.entity.tech.ChemicalGeneratorBlockEntity
import ru.dbotthepony.mc.otm.capability.energy.BlockEnergyStorageImpl
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.defineDecimal
import ru.dbotthepony.mc.otm.registry.MNames

object MachinesConfig : AbstractConfig("machines") {
	init {
		AndroidStationBlockEntity.registerConfig(builder)
		ChemicalGeneratorBlockEntity.registerConfig(builder)
		MatterBottlerBlockEntity.registerConfig(builder)
		MatterReplicatorBlockEntity.registerConfig(builder)
	}

	val PLATE_PRESS = workerValues(
		MNames.PLATE_PRESS,
		energyStorage = Decimal(40_000),
		energyThroughput = Decimal(200),
		powerConsumption = Decimal(15)
	)

	val MATTER_DECOMPOSER = workerValues(
		MNames.MATTER_DECOMPOSER,
		energyStorage = Decimal(100_000),
		energyThroughput = Decimal(400),
		powerConsumption = Decimal(240),
		matterCapacity = Decimal(400),
	)

	val MATTER_SCANNER = workerValues(
		MNames.MATTER_SCANNER,
		energyStorage = Decimal(40_000),
		energyThroughput = Decimal(200),
		powerConsumption = Decimal(80),
	)

	private val MATTER_RECYCLER = workerValues(
		MNames.MATTER_RECYCLER,
		energyStorage = Decimal(80_000),
		energyThroughput = Decimal(400),
		powerConsumption = Decimal(100),
		matterCapacity = Decimal(400),
		workTimeMultiplier = null
	) {
		MatterRecycler
	}

	object MatterRecycler {
		val VALUES by ::MATTER_RECYCLER
		val MATTER_PER_TICK by builder.defineDecimal("MATTER_PER_TICK", Decimal("0.5"))
	}

	private val MATTER_RECONSTRUCTOR = workerValues(
		MNames.MATTER_RECONSTRUCTOR,
		energyStorage = Decimal(100_000),
		energyThroughput = Decimal(1000),
		powerConsumption = Decimal(400),
		matterCapacity = Decimal(200),
		workTimeMultiplier = null,
	) {
		MatterReconstructor
	}

	object MatterReconstructor {
		val VALUES by ::MATTER_RECONSTRUCTOR

		val ALLOW_TO_SKIP_ANVIL: Boolean by builder
			.comment("Allow to repair tools without having their anvil-repair items researched")
			.comment("Ignored if onlyAnvil is true")
			.define("ALLOW_TO_SKIP_ANVIL", false)

		val ONLY_ANVIL: Boolean by builder
			.comment("Force repairing only by matter value of anvil materials")
			.comment("Doesn't make logical sense but might be good for balancing")
			.define("ONLY_ANVIL", false)

		val FAILURE_CHANCE: Double by builder
			.comment("Constant additional chance (over pattern research factor) that replication will fail at any given tick")
			.comment("In event of failure repair tick is wasted, as well as resources")
			.defineInRange("FAILURE_CHANCE", 0.01, 0.0, 0.99)

		val DIVISOR: Double by builder
			.comment("Magnitute of **slowdown** of repairer.")
			.comment("If this value is 1, repairer will repair item from 0% to 100% in COMPLEXITY (of tool) ticks")
			.comment("(or twice the complexity of anvil repair ingredients if tool itself has no matter value)")
			.comment("---")
			.comment("If value is smaller than 1, repairer will repair items faster")
			.comment("If value is bigger than 1, repairer will repair items slower")
			.defineInRange("DIVISOR", 3.0, 0.1, Double.MAX_VALUE)
	}

	val STORAGE_POWER_SUPPLIER = BlockEnergyStorageImpl.makeConfigEntry(builder, MNames.STORAGE_POWER_SUPPLIER, capacity = Decimal(100_000), throughput = Decimal(320))
	val STORAGE_INTERFACES = BlockEnergyStorageImpl.makeConfigEntry(builder, "STORAGE_INTERFACES", capacity = Decimal(10_000))
	val ITEM_MONITOR = BlockEnergyStorageImpl.makeConfigEntry(builder, MNames.ITEM_MONITOR)
	val DRIVE_VIEWER = BlockEnergyStorageImpl.makeConfigEntry(builder, MNames.DRIVE_VIEWER)
	val DRIVE_RACK = BlockEnergyStorageImpl.makeConfigEntry(builder, MNames.DRIVE_RACK, capacity = Decimal(80_000))

	object AndroidCharger {
		val RADIUS_WIDTH: Double by builder
			.comment("Effective charger range on horizontal plane")
			.defineInRange("RADIUS_WIDTH", 16.0, 1.0, Double.MAX_VALUE)

		val RADIUS_HEIGHT: Double by builder
			.comment("Effective charger range on vertical plane")
			.defineInRange("RADIUS_HEIGHT", 4.0, 1.0, Double.MAX_VALUE)
	}

	val ANDROID_CHARGER = BlockEnergyStorageImpl.makeConfigEntry(builder, MNames.ANDROID_CHARGER, capacity = Decimal(1_000_000), throughput = Decimal(8192)) { AndroidCharger }

	val POWERED_FURNACE = workerValues(
		"POWERED_FURNACE",
		energyStorage = Decimal(40_000),
		energyThroughput = Decimal(200),
		powerConsumption = Decimal(20),
		workTimeMultiplier = 0.75
	)

	val POWERED_BLAST_FURNACE = workerValues(
		"POWERED_BLAST_FURNACE",
		energyStorage = Decimal(40_000),
		energyThroughput = Decimal(200),
		powerConsumption = Decimal(20),
		workTimeMultiplier = 0.75
	)

	val POWERED_SMOKER = workerValues(
		"POWERED_SMOKER",
		energyStorage = Decimal(40_000),
		energyThroughput = Decimal(200),
		powerConsumption = Decimal(10),
		workTimeMultiplier = 0.75
	)

	object Upgrades {
		init {
			builder.push("UPGRADES")
		}

		val MIN_SPEED: Double by builder
			.comment("Minimal combined upgrade speed percentage (upgrades can't decrease machine speed below this)")
			.defineInRange("MIN_SPEED", -0.8, -1.0, Double.MAX_VALUE)

		val MAX_SPEED: Double by builder
			.comment("Maximal combined upgrade speed percentage")
			.defineInRange("MAX_SPEED", Double.MAX_VALUE, 0.0, Double.MAX_VALUE)

		val MIN_ENERGY by builder
			.comment("Minimal combined energy consumption percentage (upgrades can't decrease machine energy consumption below this)")
			.defineDecimal("MIN_ENERGY", Decimal(-0.8), Decimal(-1.0))

		val MAX_ENERGY by builder
			.comment("Maximal combined energy consumption percentage")
			.defineDecimal("MAX_ENERGY", Decimal.LONG_MAX_VALUE, Decimal.ZERO)

		init {
			builder.pop()
		}
	}

	init {
		Upgrades
	}
}
